-----------------------------------------------------------------------------------------------
-- Client Lua Script for LookAtMe
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
 
require "Window"
require "Unit"
require "ICCommLib"
require "ICComm"
require "ChatSystemLib"

-----------------------------------------------------------------------------------------------
-- LookAtMe Module Definition
-----------------------------------------------------------------------------------------------
local LookAtMe = {} 
-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
local function DistanceToUnit(unitTarget)
	--if self.bUseDistance ~= true then return nil end
	
	local unitPlayer = GameLib.GetPlayerUnit()
	if type(unitTarget) == "string" then
		unitTarget = GameLib.GetPlayerUnitByName(tostring(unitTarget))
	end
	
	if not unitTarget or not unitPlayer then
		return 0
	end

	tPosTarget = unitTarget:GetPosition()
	tPosPlayer = unitPlayer:GetPosition()

	if tPosTarget == nil or tPosPlayer == nil then
		return 0
	end

	local nDeltaX = tPosTarget.x - tPosPlayer.x
	local nDeltaY = tPosTarget.y - tPosPlayer.y
	local nDeltaZ = tPosTarget.z - tPosPlayer.z

	local nDistance = math.floor(math.sqrt((nDeltaX ^ 2) + (nDeltaY ^ 2) + (nDeltaZ ^ 2)))
	return nDistance
end

local function strsplit(sep, str)
	local sep, fields = sep or ":", {}
	local pattern = string.format("([^%s]+)", sep)
	string.gsub(str ,pattern, function(c) fields[#fields+1] = c end)
	return fields
end
-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function LookAtMe:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 
	self.tRPers = {}
	self.bIsOn = true
	self.arUnit2Nameplate = {}
	self.strEnable = "Look At Me, RP! is currently enabled"
    -- initialize variables here

    return o
end

function LookAtMe:Init()
	local bHasConfigureFunction = false
	local strConfigureButtonText = ""
	local tDependencies = {
		-- "UnitOrPackageName",
		"OneVersion",
		"ChatLog",
	}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end

 function LookAtMe:OnDependencyError(strDep, strError)
	-- if you don't care about this dependency, return true.
	if strDep == "OneVersion" then
		return true
	elseif strDep == "ChatLog" then
		return true
	end
	return false
end
-----------------------------------------------------------------------------------------------
-- LookAtMe OnLoad
-----------------------------------------------------------------------------------------------
function LookAtMe:OnLoad()
    -- load our form file
	self.xmlDoc = XmlDoc.CreateFromFile("LookAtMe.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
	Apollo.RegisterEventHandler("UnitCreated","OnUnitCreated",self) 
	Apollo.RegisterEventHandler("UnitDestroyed","OnUnitDestroyed",self)
	Apollo.RegisterEventHandler("ChangeWorld", "OnWorldChange", self)
end
-----------------------------------------------------------------------------------------------
-- LookAtMe OnDocLoaded
-----------------------------------------------------------------------------------------------
function LookAtMe:OnDocLoaded()

	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
		-- if the xmlDoc is no longer needed, you should set it to nil
		-- self.xmlDoc = nil
		-- Register handlers for events, slash commands and timer, etc.
		-- e.g. Apollo.RegisterEventHandler("KeyDown", "OnKeyDown", self)
		Apollo.RegisterSlashCommand("lookatme", "OnLookAtMeOn", self)
		self.tmrJoinChannel = ApolloTimer.Create(5.0, true, "OnJoinChannelTimer", self)
		self.tmrNamePlareRefresh = ApolloTimer.Create(2, true, "RefreshPlates", self)
		self.tmrResendRequest = ApolloTimer.Create(5, false, "Resend", self)
		-- Do additional Addon initialization here
		if self.bIsOn == false then self.strEnable = "Look At Me, RP! is currently disabled." else self.strEnable = "Look At Me, RP! is currently enabled" end
		ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_Command, self.strEnable, "" )
	end 
end
-----------------------------------------------------------------------------------------------
-- Save and Restore Data
-----------------------------------------------------------------------------------------------
function LookAtMe:OnSave(eLevel)
	-- This example uses account level saves.
	if eLevel == GameLib.CodeEnumAddonSaveLevel.Account then
	-- Set your variables into tData
		return { tRPers = self.tRPers, bIsOn = self.bIsOn}
	end
end

function LookAtMe:OnRestore(eLevel, tData)
	if eLevel == GameLib.CodeEnumAddonSaveLevel.Account then
		-- Set your reference for the saved variables
		if tData then
			self.tRPers = tData.tRPers
			self.bIsOn = tData.bIsOn or true
		end
	end
end
-----------------------------------------------------------------------------------------------
-- LookAtMe Functions
-----------------------------------------------------------------------------------------------
-- Define general functions here
function LookAtMe:HelperVerifyVisibilityOptions(tNameplate)
	local unitOwner = tNameplate.unitOwner
	local bHiddenUnit = not unitOwner:ShouldShowNamePlate()
	if self.bIsOn == false then
		return false
	end
	if bHiddenUnit then
		return false
	end
		
	if tNameplate.bOccluded or not tNameplate.bOnScreen then
		return false
	end

	return true
end

function LookAtMe:OnUnitOcclusionChanged(wndHandler, wndControl, bOccluded)
	local unitOwner = wndHandler:GetData().unitOwner
	self.arUnit2Nameplate[unitOwner:GetId()].bOccluded = bOccluded
	self:RefreshPlates(unitOwner)
end

function LookAtMe:OnWorldLocationOnScreen(wndHandler, wndControl, bOnScreen)
		local unitOwner = wndHandler:GetData().unitOwner
	if self.arUnit2Nameplate[unitOwner:GetId()] ~= nil then
		self.arUnit2Nameplate[unitOwner:GetId()].bOnScreen = bOnScreen
	end
end

function LookAtMe:RefreshPlates(unitNew)
	if type(unitNew) == "number" then return end
	if unitNew and unitNew:IsThePlayer() then return end
	if unitNew and unitNew:IsACharacter() then
		local tNameplate = self.arUnit2Nameplate[unitNew:GetId()]
		if self.bIsOn == false then
			tNameplate.wndNameplate:Show(false, false)
			tNameplate.bShow = false
		else
			local bNewShow = self:HelperVerifyVisibilityOptions(tNameplate) and (DistanceToUnit(tNameplate.unitOwner) <= 50)
			if bNewShow ~= tNameplate.bShow then
				tNameplate.wndNameplate:Show(bNewShow, false)
				tNameplate.bShow = bNewShow
			end
		end
		return
	end
	for idx, tNameplate in pairs(self.arUnit2Nameplate) do
		if self.bIsOn == false then
			tNameplate.wndNameplate:Show(false, false)
			tNameplate.bShow = false
		else
			local bNewShow = self:HelperVerifyVisibilityOptions(tNameplate) and (DistanceToUnit(tNameplate.unitOwner) <= 50)
			if bNewShow ~= tNameplate.bShow then
				tNameplate.wndNameplate:Show(bNewShow, false)
				tNameplate.bShow = bNewShow
			end
		end
	end
end

function LookAtMe:OnJoinChannelTimer()
	--Print("Join Timer triggering")
	self.chnLookAtMe = ICCommLib.JoinChannel("LookAtMe", ICCommLib.CodeEnumICCommChannelType.Global)
	if self.chnLookAtMe:IsReady() then
		--Print("Channel created, setting join result function.")
		self.chnLookAtMe:SetReceivedMessageFunction("OnMessageReceived", self)
		self.chnLookAtMe:SetSendMessageResultFunction("OnMessageSent", self)
		self.tmrJoinChannel:Stop()
	end
end

function LookAtMe:OnMessageReceived(channel, strMessage, idMessage)
	--Print("Message Received: "..strMessage)
	--"SenderName|message"
	local tMessage = strsplit("|", strMessage)
	if tMessage[2] == "true" then
		table.insert(self.tRPers,tMessage[1])
		local unitNew = GameLib.GetPlayerUnitByName(tMessage[1])
		self:CreateRPIcon(unitNew)
	elseif tMessage[2] == "Request" then
		if self.chnLookAtMe and self.chnLookAtMe:IsReady() then
			local strMessage = GameLib.GetPlayerUnit():GetName().."|true"
			local strTarget = tMessage[1]
			local idId = self.chnLookAtMe:SendPrivateMessage( strTarget, strMessage )
			table.insert(self.tMessages, idId, {strTarget, strMessage} )
		end
	end
end

function LookAtMe:OnMessageSent(iccomm, eResult, idMessage)
	if eResult == ICCommLib.CodeEnumICCommMessageResult.Sent then
		--Print("Message Sent Correctly.")
		table.remove(self.tMessages, idMessage)
	elseif eResult == ICCommLib.CodeEnumICCommMessageResult.NotInChannel then
		Apollo.AddAddonErrorText(self, "Not in Channel.")
	elseif eResult == ICCommLib.CodeEnumICCommMessageResult.Throttled then
		Apollo.AddAddonErrorText(self, "Message Throttled: "..self.tMessages[idMessage][2])
	end
end

function LookAtMe:RequestPlayerStatus(unitNew)
	if unitNew == nil then return end
	local strUnitName
	if type(unitNew) == "string" then
		strUnitName = unitNew
	elseif unitNew:IsACharacter() then
		strNew = unitNew:GetName()
	else
		return
	end
	local bIsKnown = false
	for i,v in pairs(self.tRPers) do
		if v == strUnitName then
			self:CreateRPIcon(unitNew)
			bIsKnown = true
			return
		end
	end
	if bIsKnown == false then
		if self.chnLookAtMe and self.chnLookAtMe:IsReady() then
			local strMessage =  GameLib.GetPlayerUnit():GetName().."|Request"
			local strTarget = strUnitName
			local idId = self.chnLookAtMe:SendPrivateMessage( strTarget, strMessage )
			table.insert(self.tMessages, idId, {strTarget, strMessage} )
		end
	end
end

function LookAtMe:CreateRPIcon(unitNew)
	local idUnit = unitNew:GetId()
	if self.arUnit2Nameplate[idUnit] ~= nil and self.arUnit2Nameplate[idUnit].wndNameplate:IsValid() then
		return
	end
	
	local wndNew = Apollo.LoadForm(self.xmlDoc, "LookAtMeIcon", nil, self)
	wndNew:SetUnit(unitNew, 1)
	
	local tNameplate =
	{
		unitOwner 		= unitNew,
		idUnit 			= idUnit,
		unitName		= strUnitName,
		wndNameplate	= wndNew,
		bOnScreen 		= wndNew:IsOnScreen(),
		bOccluded 		= wndNew:IsOccluded(),
		eDisposition	= unitNew:GetDispositionTo(self.unitPlayer),
		bShow			= self.bIsOn,
	}
	
	wndNew:SetData(
		{
			unitName = strUnitName,
			unitOwner = unitNew,
		}
	)
	self.arUnit2Nameplate[idUnit] = tNameplate
end

function LookAtMe:OnUnitCreated(unitNew)
	if not self.unitPlayer then
		self.unitPlayer = GameLib.GetPlayerUnit()
	end
	if unitNew:IsACharacter() then
		self:RequestPlayerStatus(unitNew)
	end
end

function LookAtMe:OnUnitDestroyed(unitOwner)
	if unitOwner:IsACharacter() then
		local idUnit = unitOwner:GetId()
		if self.arUnit2Nameplate[idUnit] == nil then
			return
		end		
		local wndNameplate = self.arUnit2Nameplate[idUnit].wndNameplate
		wndNameplate:Destroy()
		self.arUnit2Nameplate[idUnit] = nil
	end
end

function LookAtMe:OnWorldChange()
	for i, v in pairs(self.arUnit2Nameplate) do
		local wndNameplate = self.arUnit2Nameplate[i].wndNameplate
		wndNameplate:Destroy()
		self.arUnit2Nameplate[i] = nil
	end
end

-- on SlashCommand "/lookatme"
function LookAtMe:OnLookAtMeOn()
	self.bIsOn = not self.bIsOn
	self:RefreshPlates()
	if self.bIsOn == false then self.strEnable = "Look At Me, RP! is currently disabled." else self.strEnable = "Look At Me, RP! is currently enabled" end
	ChatSystemLib.PostOnChannel( ChatSystemLib.ChatChannel_Command, self.strEnable, "" )
end

-----------------------------------------------------------------------------------------------
-- LookAtMe Instance
-----------------------------------------------------------------------------------------------
local LookAtMeInst = LookAtMe:new()
LookAtMeInst:Init()
